import bpy
import math as m
import time
import os
import random as r
from array import *
from mathutils import *


pi=3.141592654

##########################################################################
# Delete Objects
##########################################################################
# Alles löschen
def DeleteAll(do=0):
	if do==0:
		bpy.ops.object.select_all(action="SELECT")
		bpy.ops.object.delete(use_global=False)
		
		for block in bpy.data.cameras:
			if block.users == 0:
				bpy.data.cameras.remove(block)
				
		for block in bpy.data.meshes:
			if block.users == 0:
				bpy.data.meshes.remove(block)

		for block in bpy.data.materials:
			if block.users == 0:
				bpy.data.materials.remove(block)

		for block in bpy.data.textures:
			if block.users == 0:
				bpy.data.textures.remove(block)

		for block in bpy.data.images:
			if block.users == 0:
				bpy.data.images.remove(block)   
				
		for block in bpy.data.curves:
			if block.users == 0:
				bpy.data.curves.remove(block)

		for block in bpy.data.lights:
			if block.users == 0:
				bpy.data.lights.remove(block)
			
	else:
		DeleteAll(0)
		context = bpy.context
		scene = context.scene

		for c in scene.collection.children:
			scene.collection.children.unlink(c)
			bpy.data.collections.remove(c)
			
			
		for c in bpy.data.collections:
			if not c.users:
				bpy.data.collections.remove(c)	

		
	
##########################################################################	
def Delete_Selected():
	bpy.ops.object.delete()
##########################################################################	
def Delete_Object_by_name(name):
	# Select the object
	bpy.data.objects[name].select = True
	bpy.ops.object.delete()
##########################################################################
def Delete_Object_by_ID(id):
	# Select the object
	bpy.data.objects[id.name].select = True
	bpy.ops.object.delete()	
	
##########################################################################
# Material
##########################################################################


##########################################################################
def Get_Material(mat_name):
    materials = bpy.data.materials
    # if .get returns None, the assignment comes from the right of the 'or'
    mat = materials.get(mat_name) or materials.new(mat_name)
    return mat		
##########################################################################	


##########################################################################
# Do with Collections
##########################################################################
#Make new Collection
def NewCollection(name):
    #Linking collections
    # Create new collection
    col = bpy.data.collections.new(name)
    # The collection hasn't been added as children of the root collection
    # Need to do this step to be visible in outline
    root_col = bpy.context.scene.collection
    root_col.children.link(col)
    return col
##########################################################################
def LinkToCollection(col,obj):
    col.objects.link(obj)
##########################################################################
def UnlinkObjFromScene(obj):
    scene = bpy.context.scene
    scene.collection.objects.unlink(obj)
##########################################################################
def MoveObjToCollection(col,obj):
    LinkToCollection(col,obj)		
    UnlinkObjFromScene(obj)	
##########################################################################
def makelinked(obx):
    myo=obx.copy()
    scene = bpy.context.scene
    scene.collection.objects.link(myo)
    return myo
##########################################################################
def Delete_Collection(name = "Collection 1"):
	remove_collection_objects = True
	coll = bpy.data.collections.get(name)
	if coll:
		if remove_collection_objects:
			obs = [o for o in coll.objects if o.users == 1]
			while obs:
				bpy.data.objects.remove(obs.pop())
	bpy.data.collections.remove(coll)

# Enumerate Collection-Objects
"""
import bpy

wcol = bpy.data.collections.get("Collection 1")
if wcol:
    for i, o in enumerate(wcol.objects):
        o.name = "Water%d" % i
"""		
	
##########################################################################

##########################################################################
# RigidBody
##########################################################################
# type='ACTIVE' bbox='MESH','CONVEX_HULL'
def AddRigidBody(type,bbox,ac=0):
    bpy.ops.rigidbody.object_add()
    bpy.context.object.rigid_body.type = type
    bpy.context.object.rigid_body.restitution = 0.2
    bpy.context.object.rigid_body.friction = 0.111354
    bpy.context.object.rigid_body.collision_shape = bbox
    RB_Deactivate(ac)
##########################################################################
def RigidBodyAdd(ap='A',typ='B'):
    if (ap=='A'):  
        sort='ACTIVE'
    else:
        sort='PASSIVE'
           
    if typ=='B':
        typ='BOX'
    elif typ=='S':
        typ='SPHERE'
    elif typ=='C':
        typ='CONVEX HULL'
    else: 
        typ='MESH'  
    AddRigidBody(sort,typ)	
##########################################################################
def RB_Deactivate(pa=1):
    if pa==1:
       val=True
    else:   
       val=False
    bpy.context.object.rigid_body.use_deactivation = val
    bpy.context.object.rigid_body.use_start_deactivated = val
##########################################################################	 
def RB_Bake(Steps,Solver):   
    bpy.ops.ptcache.free_bake_all()
    RB_SetCalculation(Steps,Solver)
    bpy.ops.ptcache.bake_all(bake=True)
    print ("OK Bake DONE")	
##########################################################################
# Zum Beispiel 240,80
def RB_SetCalculation(Steps,Solver):
    bpy.context.scene.rigidbody_world.steps_per_second = Steps
    bpy.context.scene.rigidbody_world.solver_iterations = Solver
##########################################################################
# Aktives Objekt wird FLUID-Effektor
def MF_SetEffector():
	bpy.ops.object.modifier_add(type='FLUID')
	bpy.context.object.modifiers["Fluid"].fluid_type = 'EFFECTOR'
	bpy.context.object.modifiers["Fluid"].effector_settings.subframes = 5
	bpy.context.object.modifiers["Fluid"].effector_settings.surface_distance = 0.01
	
##########################################################################
# Aktives Objekt wird FLUID-Domain
def MF_SetDomain():
	bpy.ops.object.modifier_add(type='FLUID')
	bpy.context.object.modifiers["Fluid"].fluid_type = 'DOMAIN'
	bpy.context.object.modifiers["Fluid"].domain_settings.domain_type = 'LIQUID'
	bpy.context.object.modifiers["Fluid"].domain_settings.cache_directory = "\\cache"
	bpy.context.object.modifiers["Fluid"].domain_settings.cache_type = 'FINAL'	
	bpy.context.object.modifiers["Fluid"].domain_settings.use_mesh = True
	bpy.context.object.modifiers["Fluid"].domain_settings.delete_in_obstacle = True
	bpy.context.object.modifiers["Fluid"].domain_settings.use_flip_particles = True
	
##########################################################################
def MF_SetInFlow():
	bpy.ops.object.modifier_add(type='FLUID')
	bpy.context.object.modifiers["Fluid"].fluid_type = 'FLOW'
	bpy.context.object.modifiers["Fluid"].flow_settings.flow_type = 'LIQUID'
	bpy.context.object.modifiers["Fluid"].flow_settings.flow_behavior = 'INFLOW'
	bpy.context.object.modifiers["Fluid"].flow_settings.subframes = 10
##########################################################################


##########################################################################
# Geometrical stuff
##########################################################################
#
def Setloc(myo,xp,yp,zp):
#print("Moving: ",myo.name," to ",xp,yp,zp)
	myo.location.x=xp
	myo.location.y=yp
	myo.location.z=zp
##########################################################################	
def SetRoatation(myi,xo,yp,zp):
	myo.rotation_euler.x=m.radians(xp)
	myo.rotation_euler.y=m.radians(yp)
	myo.rotation_euler.z=m.radians(zp)
##########################################################################	
def ReCenterGeo(obj,hg):	
	myobj = bpy.data.objects[obj.name]
	bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')
	myobj.location=(0,0,hg)
	return myobj
##########################################################################
# Make a Plank out of a cube
def cube2plank(obj,sz,psz):    
    # Modify Cube 1 to Plank (1,3,5,7)
    obj.data.vertices[0].co.z -= sz
    obj.data.vertices[2].co.z -= sz

    obj.data.vertices[4].co.z -= sz
    obj.data.vertices[6].co.z -= sz

    for a in range(0,4):
        obj.data.vertices[a].co.x -= psz
        obj.data.vertices[a+4].co.x += psz
##########################################################################	
"""
    Rotate obj to look at target

    :arg obj: the object to be rotated. Usually the camera
    :arg target: the location (3-tuple or Vector) to be looked at
    :arg roll: The angle of rotation about the axis from obj to target in radians. 
	
	Sampole:
	import math
	from mathutils import *
	cube = bpy.data.objects["Cube"]
	# cube.location = (5, -5, 5)
	cam = bpy.data.objects["Camera"]
	cam.location = (5, -5, -2)
	Point_at(cam, cube, roll=45)

    Based on: https://blender.stackexchange.com/a/5220/12947 (ideasman42)      
"""
##########################################################################	
def Point_at(obj, tar, roll=0):   
    target=tar.location   
    if not isinstance(target, Vector):
        target = Vector(target)
    loc = obj.location
    # direction points from the object to the target
    direction = target - loc
    
    quat = direction.to_track_quat('-Z', 'Y')

    # /usr/share/blender/scripts/addons/add_advanced_objects_menu/arrange_on_curve.py
    quat = quat.to_matrix().to_4x4()
    rollMatrix = Matrix.Rotation(m.radians(roll), 4, 'Z')

    # remember the current location, since assigning to obj.matrix_world changes it
    loc = loc.to_tuple()
    obj.matrix_world = quat @ rollMatrix
    obj.location = loc


##########################################################################		
# Speed - Speed with which the ball rotates
# Keyf - Last Keyframe		
##########################################################################
# cam = bpy.data.objects["Camera"]
def Keyframe_LR(cam,fn):
	cam.keyframe_insert(data_path="rotation_euler", frame=fn)
	cam.keyframe_insert(data_path="location", frame=fn)
##########################################################################
def Camera_Keyframe_LR(fn):
	#cam = bpy.data.objects["Camera"]
	cam=Get_Active_Cam()
	Keyframe_LR(cam,fn)
##########################################################################
def Camera_Setloc(xp,yp,zp):	
	#myo= bpy.data.objects["Camera"]
	myo=Get_Active_Cam()
	Setloc(myo,xp,yp,zp)
##########################################################################
def Get_Active_Cam():
	cam = bpy.context.scene.camera
	if cam is None:
		Cameras=NewCollection("Cameras")
		cam=AddCamera("Camera")
		MoveObjToCollection(Cameras,cam)
		#print("no scene camera in Get_Active_Cam()")		
	return(cam)	
##########################################################################
def Set_Active_Cam():
	scene = bpy.context.scene
	currentcam = bpy.context.scene.camera
	setcam = False

	for ob in scene.objects:
		if ob.type == 'CAMERA':
			if ob == currentcam:
				setcam = True
			elif setcam:
				bpy.context.scene.camera = ob
				break

	if currentcam == bpy.context.scene.camera:      
		for ob in scene.objects:
			if ob.type == 'CAMERA':
				bpy.context.scene.camera = ob
				break
##########################################################################	
def Set_Active_Cam_ID(ob):	
	bpy.context.scene.camera = ob
##########################################################################	
def Set_Active_Cam_Name(Name='CAMERA'):	
	for ob in scene.objects:
			if ob.name == Name:
				bpy.context.scene.camera = ob
				break
##########################################################################
# X=0, Y=1, Z=2
def Active_Camera_turn_Any(grad,dir=0):	
	cam=Get_Active_Cam()
	Camera_turn_Any(cam,grad,dir)
##########################################################################
def Active_Camera_turn_X(Grad):
	Active_Camera_turn_Any(Grad,0)
##########################################################################
def Active_Camera_turn_Y(Grad):
	Active_Camera_turn_Any(Grad,1)
##########################################################################
def Active_Camera_turn_Z(Grad):
	Active_Camera_turn_Any(Grad,2)	
##########################################################################
# Grad - Winkel der Drehung
# dir: X=0, Y=1, Z=2
# cam - ID of Camera 
def Camera_turn_Any(cam,grad=45,dir=0):		
	Euler=m.radians(grad)
	yrot=cam.rotation_euler[dir]
	cam.rotation_euler[dir] = yrot+Euler
##########################################################################
def Camera_turn_X(cam,Grad=45):
	Camera_turn_Any(cam,Grad,0)
##########################################################################
def Camera_turn_Y(cam,Grad=45):
	Camera_turn_Any(cam,Grad,1)
##########################################################################
def Camera_turn_Z(cam,Grad=45):
	Camera_turn_Any(cam,Grad,2)	
##########################################################################
# Verschiebe Objekte in X-Richtung
def Move_selected_ObjectX(va):
	for obj in bpy.context.selected_objects:
		obj.location.x+=va	
##########################################################################
# Verschiebe Objekte in Y-Richtung
def Move_selectd_ObjectY(va):
	for obj in bpy.context.selected_objects:
		obj.location.y+=va	
##########################################################################	
# Verschiebe Objekte in Z-Richtung
def Move_selectd_ObjectZ(va):
	for obj in bpy.context.selected_objects:
		obj.location.z+=va
##########################################################################	
# Setze Position von selektierten Objekten in X-Richtung
def Move_selected_Object_to_X(va):
	for obj in bpy.context.selected_objects:
		obj.location.x=va	
##########################################################################
# Setze Position von selektierten Objekten in Y-Richtung
def Move_selected_Object_to_Y(va):
	for obj in bpy.context.selected_objects:
		obj.location.y=va	
##########################################################################
# Setze Position von selektierten Objekten in Z-Richtung
def Move_selected_Object_to_Z(va):
	for obj in bpy.context.selected_objects:
		obj.location.z=va	
##########################################################################
# Drehe selektierte Objekte in X-Richtung um va Grad
def Rotate_selected_ObjectX(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.x += math.radians(va)	
##########################################################################
# Drehe selektierte Objekte in Y-Richtung um va Grad
def Rotate_selected_ObjectY(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.y += math.radians(va)
##########################################################################
# Drehe selektierte Objekte in Z-Richtung um va Grad
def Rotate_selected_ObjectZ(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.z += math.radians(va)
##########################################################################
# Drehe selektierte Objekte in X-Richtung auf va Grad
def Set_Rotate_selected_ObjectX(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.x = math.radians(va)	
##########################################################################
# Drehe selektierte Objekte in Y-Richtung auf va Grad
def Set_Rotate_selected_ObjectY(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.y = math.radians(va)	
##########################################################################
# Drehe selektierte Objekte in Z-Richtung auf va Grad
def Set_Rotate_selected_ObjectZ(va):
	for obj in bpy.context.selected_objects:
		obj.rotation_euler.z = math.radians(va)	
##########################################################################



##########################################################################

def Euler_to_Grad(Euler):
	a=float(Euler)
	Grad=m.degrees(a)
	return(Grad)
##########################################################################	
def Grad_to_Eulerx(Grad):
    a=float(Grad)
    Euler=m.radians(a)
    return(Euler)	
##########################################################################	
"""
def merge_objects( objects, name='_temp_', transform=None ):
    assert objects
    copies = []
    for ob in objects:
        ob.select = False
        if ob.type == 'MESH':
            o2 = ob.copy(); copies.append( o2 )
            o2.data = o2.to_mesh(bpy.context.scene, True, "PREVIEW") # collaspe modifiers
            while o2.modifiers:
                o2.modifiers.remove( o2.modifiers[0] )
            if transform:
                o2.matrix_world =  transform * o2.matrix_local
            bpy.context.scene.objects.link( o2 ) #; o2.select = True
    merged = merge( copies )
    merged.name = name
    merged.data.name = name
    return merged
##########################################################################
def merge( objects ):
    print('MERGE', objects)
    for ob in bpy.context.selected_objects:
        ob.select = False
    for ob in objects:
        print('\t'+ob.name)
        ob.select = True
        assert not ob.library
    bpy.context.scene.objects.active = ob
    bpy.ops.object.join()
    return bpy.context.active_object	
##########################################################################
def gather_instances():
    instances = {}
    for ob in bpy.context.scene.objects:
        if ob.data and ob.data.users > 1:
            if ob.data not in instances:
                instances[ ob.data ] = []
            instances[ ob.data ].append( ob )
    return instances
"""	
##########################################################################def	
def Get_active_Objects():
	act=bpy.context.selected_objects
	return act	
##########################################################################def  
def Get_Active_Scene():	
	active=bpy.context.screen.scene
	return active
##########################################################################
def Get_Active_Object():
	active=bpy.context.active_object
	return active
##########################################################################
def Get_Last_Object():
	active=obj=bpy.context.selected_objects[0]
	return active
##########################################################################
def Get_Active_Material():
	active=bpy.context.scene.node_tree.nodes.active
	return active
##########################################################################
# Set Shading for active Object
# Flat Shading: False, SMooth Shading - True
def Set_Shading(Pa=True):
	bpy.context.object.data.polygons.foreach_set('use_smooth',  [Pa] * len(bpy.context.object.data.polygons))
	bpy.context.object.data.update()

##########################################################################	
def Get_Active_Window():
	active=bpy.context.area.spaces.active	
	return active
	
##########################################################################
def RandRotate(Speed,Keyf):
	mkf=Keyf/2
	nkf=Keyf/8
	for newBall in bpy.context.selected_objects:
		p1=nkf*2
		p2=mkf+p1
		mymkf=r.uniform(p1,p2)
		newBall.rotation_euler = (0, 0, 0) 
		newBall.keyframe_insert(data_path="rotation_euler", frame=1)
		xr = r.randint(0,Speed*2)-Speed
		yr = r.randint(0,Speed*2)-Speed
		zr = r.randint(0,Speed*2)-Speed
		newBall.rotation_euler = (xr,yr,zr)		
		newBall.keyframe_insert(data_path="rotation_euler", frame=mymkf)
		newBall.rotation_euler = (0, 0, 0)
		newBall.keyframe_insert(data_path="rotation_euler", frame=Keyf)
		
##########################################################################
def RandMove(Speed,Keyf):
	mkf=Keyf/2
	nkf=Keyf/8
	for newBall in bpy.context.selected_objects:
		p1=nkf*2
		p2=mkf+p1
		mymkf=r.uniform(p1,p2)
		
		xol=newBall.location.x
		yol=newBall.location.y
		zol=newBall.location.z
		
		newBall.keyframe_insert(data_path="location", frame=1)
		
		xr =xol+ r.randint(0,Speed*2)-Speed
		yr =yol+ r.randint(0,Speed*2)-Speed
		zr =zol+ r.randint(0,Speed*2)-Speed
		newBall.location = (xr,yr,zr)		
		newBall.keyframe_insert(data_path="location", frame=mymkf)
		newBall.location = (xol,yol,zol)
		newBall.keyframe_insert(data_path="location", frame=Keyf)
		
##########################################################################
# Gibt Objecten eine zufällige Größe zwischen Min und Max
def RandSize(Min,Max):
	for newBall in bpy.context.selected_objects:
		sz = r.uniform(Min,Max)
		newBall.scale = (sz,sz,sz)	

##########################################################################
# Add Basic Objects
##########################################################################
def AddCamera(xn="Camera"):
	camdata = bpy.data.cameras.new(name=xn)
	a = bpy.data.objects.new(name=xn, object_data=camdata)
	bpy.context.scene.collection.objects.link(a)
	#SelectObject(a)	
	return a
# select: bpy.data.objects[a.name].select = True
##########################################################################
# xn = Object-ID 
def SelectObject(xn):
	bpy.data.objects[xn.name].select = True
	
##########################################################################
# Add Plane
def AddPlane(x,y,z,siz):
    bpy.ops.mesh.primitive_plane_add(size=siz, enter_editmode=False, location=(x,y,z))
    RigidBodyAdd('P','B')
##########################################################################
# Add Cube
def AddCube(x,y,z,siz):
    bpy.ops.mesh.primitive_cube_add(size=siz,location=(x,y,z))
    RigidBodyAdd('A','B')
##########################################################################
# Add Sphere
def AddSphere(x,y,z,siz):
    bpy.ops.mesh.primitive_uv_sphere_add(radius=siz, enter_editmode=False, location=(x,y,z))
    RigidBodyAdd('A','S')
##########################################################################
# Add Menger Cube Level bis max. 4	
def AddMenger(x,y,z,siz,lev):
     bpy.ops.mesh.menger_sponge_add(align='WORLD', location=(x,y,z), rotation=(0, 0, 0), level=lev, radius=siz)
     RigidBodyAdd('A','B')	
##########################################################################
def AddRock(x,y,z,siz,lev): 	 
    bpy.ops.mesh.add_mesh_rock(preset_values='1', num_of_rocks=1, scale_X=(0.5, 1.25), skew_X=-0.5, scale_Y=(0.5, 1.25), skew_Y=-0.5, scale_Z=(0.5, 1.25), skew_Z=-0.5, use_scale_dis=True, scale_fac=(1,1,1), deform=3, rough=2, detail=2, display_detail=2, smooth_fac=2, smooth_it=2, use_generate=True, use_random_seed=True, user_seed=1)
    bpy.ops.object.apply_all_modifiers()
    RigidBodyAdd('A','M')
     
##########################################################################
def MakeNewPlank(col,Plank,xpo,ypo,zpo,rox):
# Erzeuge Linked Plank von Original
        obj=makelinked(Plank)		
        MoveObjToCollection(col,obj)
        Setloc(obj,xpo,ypo,zpo)                   
        obj.rotation_mode = 'XYZ'
        obj.rotation_euler=(0,0,rox)    
        return obj
    
##########################################################################
def sin(n):
    res=m.sin(n)
    return res
##########################################################################
def cos(n):
    res=m.cos(n)
    return res	

##########################################################################
# col - Collection where to insert the Layer
# Plank - link to Object that will be duplicated for the Planks
# si - No of Plank per Line
# cor - corners, Number of corners (Steps) can be 4 or 8
# x,y,z - Coordinates where the Tower is placed
def MakeQuad(col,Plank,si,cor,x,y,z):        
   phi=.9
   pi=3.1415926
   pi2=2*pi
   con=180/pi
   step=pi2/cor
   wi=(360/cor)/con
   i=0
   ro=0
   target=pi2-.1
   while (i<target):
    if z & 1:
      rox=43
      row=1
    else:
      rox=0
      row=0   
    xp=sin(i+rox)*si+x 
    yp=cos(i+rox)*si+y
    zp=z*(2*phi)-phi
    MakeNewPlank(col,Plank,xp,yp,zp,ro+row)
    #print ("Plank at",xp,yp,"with:",ro)
    ro+=wi
    i+=step
   
######################################################################
#
#
######################################################################   
def AddVir(number=5,size=1,dist=0.25,x=0,y=0,z=0,pow=250,ac=0,rest=0.95):
    if z==0:
       z=size*1.1 
    AddArray(0,number,dist)
    bpy.ops.object.apply_all_modifiers()
    AddArray(1,number,dist)
    bpy.ops.object.apply_all_modifiers()
    AddArray(2,number,dist)
    bpy.ops.object.apply_all_modifiers()
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.separate(type='LOOSE')
    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')   
	
##########################################################################		
# CubeSize
# 
##########################################################################   
def PlaceInCube(CubeSize=10):   
	d=int(CubeSize/2)          # size of area where the balls spawn (STRG+Q to see what I mean)
	xpos=0
	zpos=0	
	ypos=3		
	for newBall in bpy.context.selected_objects:
		xpos=r.randint(0,d)
		xpos=xpos*2-d
		ypos=r.randint(0,d)
		ypos=ypos*2-d				
		zpos=r.randint(0,d)
		zpos=zpos*2-d		
		#print('Moving Object')
		newBall.rotation_euler = (0, 0, 0)		
		newBall.location = (xpos, ypos, zpos)
		
 	
##########################################################################
# Tower_high, how many "floors" the tower shall have (the more floors the higher the Baking Precision must be set)
# cor - Corners, can be 4 or 8 (you can choose others but this will not bne stable)
# x,y,z - Coordinates where the Tower shall be built
##########################################################################

def MakeQuadTower(Tower_high,cor,x,y,z):
    #Time before the operations start
    then = time.time() 
# Add Cube with size .8
    Plank_Len=12
    siz=.8
    hg=0.9
    AddCube(0,0,0,1)
#----------------------------------------------------------------
    obj=bpy.context.selected_objects[0]
    cube2plank(obj,siz,Plank_Len)	
#----------------------------------------------------------------
# ReCenter Obj Geometry
    myobj=ReCenterGeo(obj,hg)  
    AssignNewMaterial(myobj)
    col=NewCollection("Quad")
    LinkToCollection(col,obj)
    
    si=Plank_Len+(siz*2.6)
    for i in range(1,Tower_high):
        MakeQuad(col,myobj,si,cor,x,y,i)	
    now = time.time() #Time after it finished
    print("Built Quad-Tower in: ", now-then, " seconds.") 	
##########################################################################    
# col - collection where to insert the Planks
# Plank - object/Plank that is duplicated
# ap - number of Planks in a floor /4
# xof,yof - offset for x and y (z is below)
# y - this is the z-Position
# Creates the Planks that are 90 Degree rotated
def MakeCircle_S(col,Plank,ap,xof,yof,y):
    pi=3.14159267
    pi2=2*pi
    tar=pi2-.001
    sz=.8
	# phi = Plankenhöhe (0.9)
    phi=.9
    # Plank-Length
    Plank_Len=6
    ps=6
    # Schleifen Variable für Kreis des Turmes
    i=0
    sg=1-sz 

    # Anzahl der Planks in einer Ebene z.b. 12     
    Anz_Planks=ap*4       

    # Schleifen-Increment pro Durchlauf
    inc=pi2/Anz_Planks

    # Rotations-Increment damit Planks in Kreisform sind
    ai=360/Anz_Planks
    # Das ist bei einer Plank-Lenght von 6 eben 60-3 also eigentlich Länge+10 Minus 10% damit die Planken sich nicht berühren    
    roi=-ai/((ps*10.05)-(ps/2))
        
    # Größenfaktor für Kreis        
    s=Anz_Planks*pi*.65 # 2.132 

    s=(s/100*(49/2))
    
    ro=0
    cnt=0

# Schleife die den Kreis aus Planks erstellt
    while i<tar:
    #--------------------------------------------------------------
    # Loop increment
        i+=inc
        cnt+=1
    #--------------------------------------------------------------    
# Setze Location der Plank   
        dx = sin(i)*s+xof
        dy = cos(i)*s+yof
        dz = y*(2*phi)-phi
    #--------------------------------------------------------------
    # Rotiere die Plank         
        ro+=roi    
        rox=ro+29.8

#--------------------------------------------------------------
# Erzeuge Linked Plank von Original bei xpo,ypo,zpo und Rotiere um rox
        obj=MakeNewPlank(col,Plank,dx,dy,dz,rox) 
######################################################################
# col - collection where to insert the Planks
# Plank - object/Plank that is duplicated
# ap - number of Planks in a floor /4
# xof,yof - offset for x and y (z is below)
# y - this is the z-Position
# sig - signal that will influence the Circel-Size in a way taht it all fits together
# Planks that are not rotated (follow along circle)
def MakeCircle_L(col,Plank,ap,xof,yof,y,sig):
    pi=3.1415926
    pi2=2*pi
    tar=pi2-.001
    sz=0.8
    # phi = Plankenhöhe (0.9)
    phi=0.9
    # Plank-Length
    Plank_Len=6
    ps=6
    # Schleifen Variable für Kreis des Turmes
    i=0
    sg=1-sz 

    # Anzahl der Planks in einer Ebene z.b. 12 
    Anz_Planks=ap         

    ma=Anz_Planks+1
    # Schleifen-Increment pro Durchlauf
    inc=pi2/Anz_Planks

    # Rotations-Increment damit Planks in Kreisform sind
    ai=360/Anz_Planks
    # Das ist bei einer Plank-Lenght von 6 eben 60-3 also eigentlich Länge+10 Minus 10% damit die Planken sich nicht berühren    
    roi=-ai/((ps*10.05)-(ps/2))
        
    # Größenfaktor für Kreis sif=.335-> Mitte
    saf=(1/Anz_Planks)*.25
    if sig==1:
        sif=.334-saf
    else:	
        sif=.337+saf
        
    s=Anz_Planks*pi2*sif
    
    ro=0
    cnt=0

# Schleife die den Kreis aus Planks erstellt
    while i<tar:
    #--------------------------------------------------------------
    # Loop increment
        i+=inc
        cnt+=1
    #--------------------------------------------------------------    
# Setze Location der Plank   
        dx = sin(i)*s+xof
        dy = cos(i)*s+yof
        dz = y*(2*phi)-phi
       
    #--------------------------------------------------------------
    # Rotiere die Plank         
        ro+=roi
        rox=ro+.3

#--------------------------------------------------------------
# Erzeuge Linked Plank von Original bei xpo,ypo,zpo und Rotiere um rox
        obj=MakeNewPlank(col,Plank,dx,dy,dz,rox)		
        
##########################################################################
# Will return the highest Planks Position
# x,y - Position of the Tower
# size - number of Planks for the Circle of the Tower
# hi - How High the Tower is (in floors, each Floor is 1 plank high)
def maketower(x,y,size,hi):
    then = time.time() #Time before the operations start
    # sz - Breite,12 -Anzahl Planks,xoffset,yoffset,Höhe
    Plank_Len=6
    # Add Cube with size .8
    siz=.8
    hg=0.9
    AddCube(0,0,0,1)
#----------------------------------------------------------------
    obj=bpy.context.selected_objects[0]
    cube2plank(obj,siz,Plank_Len)
#----------------------------------------------------------------
    # ReCenter Obj Geometry
    myobj=ReCenterGeo(obj,hg)
    Setloc(myobj,0,0,-10)
    #AssignNewMaterial(myobj)
    col=NewCollection("Tower")
    LinkToCollection(col,obj)    

    for za in range(1,hi):
        #ypos=m.cos(za)+y
        ypos=y
       # xpos=m.sin(za)+x
        xpos=x
        SizeA=size
        SizeB=size-(size/100*5)
        if za & 1:
            MakeCircle_S(col,myobj,SizeA,xpos,ypos,za)
        else:	
            MakeCircle_L(col,myobj,SizeB,xpos,ypos,za,1)        
            MakeCircle_L(col,myobj,SizeB,xpos,ypos,za,0)
        
    now = time.time() #Time after it finished
    print("It took: ", now-then, " seconds")
    return za

##########################################################################
# Object must be seleceted
def AddArray(num,anz,dist):
    bpy.ops.object.modifier_add(type='ARRAY')
    bpy.context.object.modifiers["Array"].relative_offset_displace[num] = dist
    if num==0:
        a=1
        b=2        
    elif num==1:
        a=0
        b=2
    elif num==2:
        a=0
        b=1
        
    bpy.context.object.modifiers["Array"].relative_offset_displace[a] = 0
    bpy.context.object.modifiers["Array"].relative_offset_displace[b] = 0 
        
    bpy.context.object.modifiers["Array"].count = anz
    bpy.context.object.modifiers["Array"].show_expanded = False
##########################################################################
##########################################################################
# number ... 12
# size ... 1
# dist ... 0.75 (must be < size to explode)
# typ - '1 - Sphere, 0 - Menger, 2- Cube, 3 - Rock
# optional parameters:
# x,y,z - where to place the Object
# pow -> 0.01 bis 10000 - Weight of objects (Power to destroy) 
# ac - 0/1 - 1-> start deactivated 0 - not deactivated
# rest - Set Bouncing Factor for Rigib Body 0 ... 1 (please set Floor to same)
def AddBomb(number=5,size=1,dist=0.25,tp=1,x=0,y=0,z=0,pow=250,ac=0,rest=0.95):
	if z==0:
	   z=size*1.1 
	if tp==0:
		AddMenger(x,y,z,size,3) # 3 is level can be 1 to 4
	elif tp==1:
		AddSphere(x,y,z,size)
	elif tp==2:
		AddCube(x,y,z,size)
	else: 
		AddRock(x,y,z,size,3)
	RB_Deactivate(ac)
	bpy.context.object.rigid_body.restitution = rest
	bpy.context.object.rigid_body.mass = pow
	AddArray(0,number,dist)
	bpy.ops.object.apply_all_modifiers()
	AddArray(1,number,dist)
	bpy.ops.object.apply_all_modifiers()
	AddArray(2,number,dist)
	bpy.ops.object.apply_all_modifiers()
	bpy.ops.object.mode_set(mode='EDIT')
	bpy.ops.mesh.separate(type='LOOSE')
	bpy.ops.object.mode_set(mode='OBJECT')    
	bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')
	
##########################################################################
# Labyrinth
##########################################################################
# bis maximal 40x40 BLocks
def MakeLabyrinth(Size=40):
	WIDTH=Size
	HEIGHT=Size
	N=1
	S=2
	E=4
	W=8

	odir = {}
	odir[N] = S
	odir[S] = N
	odir[W] = E
	odir[E] = W

	coords = []
	faces = []

	d = [[0 for col in range(WIDTH)] for row in range(HEIGHT)]

	for i in range(0,WIDTH):
		for j in range(0, HEIGHT):
			d[i][j] = 0

	def carve( x, y ):
		dirs = [N,S,W,E]
		shuffle(dirs)

		for i in dirs:
			print(i)
			nx = x
			if i == E:
				nx =  nx + 1
			if i == W:
				nx = nx - 1
			ny = y
			if i == S:
				ny = ny +1
			if i == N:
				ny = ny - 1
			if nx >= 0 and nx < WIDTH and ny >=0 and ny < HEIGHT and d[nx][ny] == 0:
					d[x][y] = d[x][y] | i
					d[nx][ny] = d[nx][ny]  | odir[i]
					carve(nx, ny)

	carve(0,0)

	idx=0
	step = 0.1

	for x in range(0,WIDTH):
		for y in range( 0, HEIGHT ):
			if ( d[x][y] & N ) == 0:
				coords.append((x * step , y * step, 0 ))
				coords.append(((x+1) * step, y * step, 0 ))
				coords.append(((x+1) * step, y * step, step ))
				coords.append((x * step, y * step, step ))
				faces.append(( idx, idx+1, idx+2, idx+3))
				idx+=4
			if ( d[x][y] & S ) == 0:
				coords.append((x * step , (y+1) * step, 0 ))
				coords.append(((x+1) * step, (y+1) * step, 0 ))
				coords.append(((x+1) * step, (y+1) * step, step ))
				coords.append((x * step, (y+1) * step, step ))
				faces.append(( idx, idx+1, idx+2, idx+3))
				idx+=4
			if ( d[x][y] & E ) == 0:
				coords.append(((x+1) * step, y * step, 0 ))
				coords.append(((x+1) * step, (y+1) * step, 0 ))
				coords.append(((x+1) * step, (y+1) * step, step ))
				coords.append(((x+1) * step, y * step, step ))
				faces.append(( idx, idx+1, idx+2, idx+3))
				idx+=4
			if ( d[x][y] & W ) == 0:
				coords.append((x * step, y * step, 0 ))
				coords.append((x * step, (y+1) * step, 0 ))
				coords.append((x * step, (y+1) * step, step ))
				coords.append((x * step, y * step, step ))
				faces.append(( idx, idx+1, idx+2, idx+3))
				idx+=4

	mesh = bpy.data.meshes.new(name="MyMesh")

	object = bpy.data.objects.new( 'MESH', mesh )
	bpy.context.collection.objects.link( object )

	mesh.from_pydata( coords, [], faces )
	mesh.update( calc_edges=True )

